#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/wait.h>

static const char* progname;

static int childProcess(int fd, const char* slave) {
	int pts;
	pid_t mypid;
	pid_t slavepid;

	mypid = getpid();
	printf("%u: Child running\n", mypid);
	pts = open(slave, O_RDWR);
	if (pts == -1) {
		fprintf(stderr, "%u: Cannot open %s: %s\n", mypid, slave, strerror(errno));
		close(fd);
		return 50;
	}
	slavepid = fork();
	if (slavepid == -1) {
		fprintf(stderr, "%u: Cannot create reader: %s\n", mypid, strerror(errno));
		close(pts);
		close(fd);
		return 51;
	}
	if (slavepid != 0) {
		int tmp;

		tmp = pts; pts = fd; fd = tmp;
	}
	while (1) {
		char buf;
		int rd;

		rd = read(pts, &buf, 1);
		if (rd == 1) {
			int wr;
			
			wr = write(fd, &buf, 1);
			if (wr == 0 || (wr == -1 && errno == EPIPE)) {
				fprintf(stderr, "%u: Write failed\n", mypid);
				break;
			}
		} else if (rd == 0) {
			break;
		} else if (errno == EPIPE) {
			fprintf(stderr, "%u: Got EPIPE\n", mypid);
			break;
		}
		if (slavepid != 0 && kill(slavepid, 0) != 0) {
			fprintf(stderr, "%u: Slave %u disappeared\n", mypid, slavepid);
			break;
		}
	}
	close(pts);
	close(fd);
	if (slavepid != 0) {
		kill(slavepid, SIGTERM);
	} else {
		kill(getppid(), SIGTERM);
	}
	return 0;
}

static int bindSocket(int unixfd, const struct sockaddr_un* sun, socklen_t sunlen) {
	struct stat stb;

	if (bind(unixfd, (struct sockaddr*)sun, sunlen) == 0) {
		return 0;
	}
	if (errno != EADDRINUSE) {
		return -1;
	}
	if (stat(sun->sun_path, &stb) == 0 && S_ISSOCK(stb.st_mode)) {
		unlink(sun->sun_path);
		return bind(unixfd, (struct sockaddr*)sun, sunlen);
	}
	errno = EADDRINUSE;
	return -1;
}

static void sigchld(int dummy) {
	while (1) {
		pid_t p;
		int status;

		p = waitpid(-1, &status, WNOHANG);
		if (p == 0) {
			return;
		}
		if (p > 0) {
			if (WIFEXITED(status)) {
				printf("Child %u exited with status %u\n", p, WEXITSTATUS(status));
			} else if (WIFSIGNALED(status)) {
				printf("Child %u was killed with signal %u\n", p, WTERMSIG(status));
			} else {
				printf("Child %u terminated with unexpected status %08X\n", p, status);
			}
			continue;
		}
		if (errno != EINTR) {
			return;
		}
	}
}

int main(int argc, char* argv[]) {
	socklen_t sunlen;
	struct sockaddr_un* sun;
	size_t vmplen;
	const char* vmp;
	const char* slave;
	int unixfd;

	progname = strrchr(argv[0], '/');
	if (progname) {
		progname++;
	} else {
		progname = argv[0];
	}

	if (argc != 3) {
		fprintf(stderr, "usage: %s <vmwarePipeName> <slaveDevice>\n",
				progname);
		return 99;
	}
	vmp = argv[1];
	slave = argv[2];

	if (signal(SIGCHLD, sigchld) == SIG_ERR) {
		fprintf(stderr, "Cannot install SIGCHLD handler: %s\n", strerror(errno));
		return 94;
	}
	umask(0);	/* Created socket should be accessible by everybody... */

	vmplen = strlen(vmp);
	sun = malloc(offsetof(struct sockaddr_un, sun_path) + vmplen + 1);
	if (!sun) {
		fprintf(stderr, "Not enough memory!\n");
		return 98;
	}
	sun->sun_family = AF_UNIX;
	memcpy(sun->sun_path, vmp, vmplen + 1);
	sunlen = SUN_LEN(sun);

	unixfd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (unixfd == -1) {
		fprintf(stderr, "Cannot create unix socket: %s\n", strerror(errno));
		return 97;
	}
	if (bindSocket(unixfd, sun, sunlen)) {
		fprintf(stderr, "Cannot bind to %s: %s\n", vmp, strerror(errno));
		close(unixfd);
		return 96;
	}
	if (listen(unixfd, 5)) {
		fprintf(stderr, "Cannot listen on %s: %s\n", vmp, strerror(errno));
		close(unixfd);
		unlink(sun->sun_path);
		return 95;
	}
	while (1) {
		int fd;

		fd = accept(unixfd, NULL, NULL);
		if (fd == -1) {
			fprintf(stderr, "Warning: accept() failed: %s\n", strerror(errno));
		} else {
			pid_t chld;

			chld = fork();
			if (chld < 0) {
				fprintf(stderr, "fork() failed: %s\n", strerror(errno));
				close(fd);
			} else if (chld == 0) {
				exit(childProcess(fd, slave));
			} else {
				printf("Accepted connection (%d-%d)\n", fd, chld);
				close(fd);
			}
		}
	}
	close(unixfd);
	return 0;
}


